/* Chapter 14. atouMT.
	Multithreaded ASCII to Unicode file conversion.
	Use threads as an alternative to overlapped I/O. 
	The resulting code is MUCH simpler. */
/* atouMT file1 file2 */
#include "EvryThng.h"

#define MAX_OVRLP 4  /* Number of worker threads */
#define RECORD_SIZE 0x1000  /* 4096 block size - Input file size should be
                            * a multiple of MAX_OVRLP * RECORD_SIZE 
                            * Increase this size for better performance - 65K is good */
#define SFP_ERROR 0xFFFFFFFF  /* Error getting size of file  */

DWORD WINAPI ReadWrite (LPVOID);
typedef struct TH_ARG_T {
	volatile HANDLE hIn;
	volatile HANDLE hOut;
	volatile DWORD ThreadNum;
	volatile LARGE_INTEGER FileSize;
} TH_ARG;

int _tmain (int argc, LPTSTR argv [])
{
	HANDLE hThr [MAX_OVRLP];
	HANDLE hProc = GetCurrentProcess ();
	DWORD i, ThId, HighPart;
	TH_ARG ThArg[MAX_OVRLP];

	printf ("WARNING, Sept 18, 2004. Use at your own risk. Potential Defects under investigation\n");
	if (argc != 3)
		ReportError (_T ("Usage: atouMT file1 file2"), 1, FALSE);

	for (i = 0; i < MAX_OVRLP; i++) {
		ThArg[i].ThreadNum = i;

		/* Note: If you create just two handles (one for input, one for output)
			and then use DuplicateHandle to create individual handles for each
			thread (i.e., ThArg[i].hOut, etc. are generated by DuplicateHandle),
			you'll have problems. It appears is if ReadFile and WriteFile are
			not thread safe when dealing with duplcated handles. This is the case
			even if you protect the ReadFile and WriteFile with a CRITIAL_SECTION
			and even if you use the no buffering and write through flags. 
			As an exercise, try using duplecated handles. You will see errors if
			you compare the unicode output file with one generated with a correct
			atouXX program. */
		/*  WARNING - Sept 18, 2004. I have had problems with this on Windows XP.
		    It appears that using multiple handles on the same file is not thread 
			safe. I get a variety of errors, such as the input handle not valid, the
			output file not valid, or, sometimes, the program runs to termination
			but the output is not correct.
			This occurs regardless of whether SetFilePointer or an overlapped
			structure is used for the file positioning.
			And. sometimes it works well. 
			I have not seen these problems before and will need to investigate
			further. USE AT YOUR OWN RISK 
		 */
		ThArg[i].hIn = CreateFile (argv [1], GENERIC_READ,
			FILE_SHARE_READ, NULL, OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL, NULL);
 		if (ThArg[i].hIn == INVALID_HANDLE_VALUE) 
			ReportError (_T ("Fatal error opening input file."), 2, TRUE);
    	ThArg[i].FileSize.LowPart = GetFileSize (ThArg[i].hIn, &HighPart);
		ThArg[i].FileSize.HighPart = HighPart;
		ThArg[i].hOut = CreateFile (argv [2], GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, CREATE_ALWAYS,
			FILE_ATTRIBUTE_NORMAL /*| FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH*/, NULL);
		/* Note: DO NOT USE THE TWO COMMENTED OUT FLAGS AS THEY REQUIRE THAT THE OUTPUT
			BUFFER LENGTHS BE SECTOR SIZE MULTIPLES (THE WRITE WILL FAIL ) */
		if (ThArg[i].hOut == INVALID_HANDLE_VALUE) 
			ReportError (_T ("Fatal error opening output file."), 3, TRUE);

		hThr [i] = (HANDLE) _beginthreadex (NULL, 0, ReadWrite, (LPVOID)&ThArg[i], 0, &ThId);
		if (hThr [i] == NULL)
			ReportError (_T ("Error creating thread."), 6, TRUE);
	}

	for (i = 0; i < MAX_OVRLP; i++) {
		WaitForSingleObject (hThr [i], INFINITE);
		CloseHandle (hThr [i]);
		CloseHandle (ThArg[i].hOut); 
		CloseHandle (ThArg[i].hIn);
	}
	_tprintf (_T ("Multithreaded ASCII to Unicode conversion completed.\n"));
	return 0;
}

DWORD WINAPI ReadWrite (LPVOID pArg)
{
  	CHAR AsRec [RECORD_SIZE];
	WCHAR UnRec [RECORD_SIZE];
	LARGE_INTEGER CurPosIn, CurPosOut;
	DWORD i, nRead = 1, nWrite, iTh;
	HANDLE hIn, hOut;
	TH_ARG * pThArg;
	OVERLAPPED OvRead = {0, 0, 0, 0, NULL}, OvWrite = {0, 0, 0, 0, NULL};

	pThArg = (TH_ARG *)pArg;

	iTh = pThArg->ThreadNum;
	hIn = pThArg->hIn; hOut = pThArg->hOut;

	CurPosIn.QuadPart = (LONGLONG) RECORD_SIZE * iTh;
	CurPosOut.QuadPart = (LONGLONG) RECORD_SIZE * (LONGLONG)(iTh * 2);
	while (CurPosIn.QuadPart < pThArg->FileSize.QuadPart) {

		OvRead.Offset      = CurPosIn.LowPart;
		OvRead.OffsetHigh  = CurPosIn.HighPart;
		OvWrite.Offset     = CurPosOut.LowPart;
		OvWrite.OffsetHigh = CurPosOut.HighPart;
		
		if (!ReadFile (hIn, AsRec, RECORD_SIZE, &nRead, &OvRead)) {
			ReportError (_T("Error reading input file"), 8, TRUE);
        }
		for (i = 0; i < nRead; i++)
			UnRec [i] = (WCHAR)AsRec [i];
		
		if (!WriteFile (hOut, UnRec, 2*nRead, &nWrite, &OvWrite)) {
			ReportError (_T("Error writing output file"), 9, TRUE);
		}
		CurPosIn.QuadPart += (LONGLONG) RECORD_SIZE * MAX_OVRLP;
		CurPosOut.QuadPart += (LONGLONG) RECORD_SIZE * 2 * MAX_OVRLP;
	}

	CloseHandle (hIn);
	CloseHandle (hOut);
	return 0;
}
