/* Chapter 15-1. Windows chmodW command (Windows-specific name to
 * distinguish from any Unix emulation system you may have installed,
 * such as cygwin. */

/* chmodW [options] mode file [GroupName]
	Update access rights of the named file.
	Options:
		-f  Force - do not complain if unable to change.
		-c  Create the file if it does not exist.
			NOT AN OPTION FOR THE STANDARD UNIX COMMAND. 
			The group name is after the file name! 
			If the group name is not given, it is taken from the
			process's token.  */
/* Requires Windows NT. This will not work under Windows 9x */

/* This program illustrates:
	1.  Setting the file security attributes.
	2.  Changing a security descriptor. */

#include "EvryThng.h"

int _tmain (int argc, LPTSTR argv [])
{
	HANDLE hFile, hSecHeap;
	BOOL Force, CreateNew, Change, Exists;
	DWORD Mode, DecMode, UsrCnt = ACCT_NAME_SIZE;
	TCHAR UsrNam [ACCT_NAME_SIZE];
	int FileIndex, GrpIndex, ModeIndex;
		/* Array of rights settings in "UNIX order". */
		/* There are distinct arrays for the allowed and denied
		 * ACE masks so that SYNCHRONIZE is never denied.
		 * This is necessary as both FILE_GENERIC_READ and FILE_GENERIC_WRITE
		 * require SYNCHRONIZE access.
		 * For more information, see:
		 * http://msdn.microsoft.com/library/default.asp?url=/library/en-us/fileio/base/file_security_and_access_rights.asp
		 */
	DWORD AllowedAceMasks [] =
		{FILE_GENERIC_READ, FILE_GENERIC_WRITE, FILE_GENERIC_EXECUTE};
	DWORD DeniedAceMasks [] =
		{FILE_GENERIC_READ & ~SYNCHRONIZE, FILE_GENERIC_WRITE & ~SYNCHRONIZE, 
									  FILE_GENERIC_EXECUTE & ~SYNCHRONIZE};
	LPSECURITY_ATTRIBUTES pSa;
	LPCTSTR GroupName = NULL;

	if (!WindowsVersionOK (3, 1)) 
		ReportError (_T("This program requires Windows NT 3.1 or greater"), 1, FALSE);

	if (argc < 3) ReportError (_T("Usage: chmodW [options] mode file [GroupName]"), 1, FALSE);
	ModeIndex = Options (argc, argv, _T ("fc"), &Force, &CreateNew, NULL);
	GrpIndex = ModeIndex + 2;
	FileIndex = ModeIndex + 1;
	DecMode = _ttoi (argv [ModeIndex]);

	/* Decimal - Assume every digit is between 0 and 7 and convert. */

	Mode = ((DecMode / 100) % 10) * 64 
			 + ((DecMode / 10) % 10) * 8 + (DecMode % 10);
	Exists = (_taccess (argv [FileIndex], 0) == 0);

	if (!Exists && CreateNew) {
				/* File does not exist; create a new one. */
		if (argc < GrpIndex) 
			ReportError (_T ("Usage: Chmod -c Perm file [GroupName]."), 1, FALSE);
		if (argc == GrpIndex) { /* Find the primary group name */
			GroupName = NULL;
		} else
			GroupName = argv[GrpIndex];

		if (!GetUserName (UsrNam, &UsrCnt))
			ReportError (_T ("Failed to get user name."), 2, TRUE);

		pSa = InitializeUnixSA (Mode, UsrNam, GroupName, AllowedAceMasks, 
				DeniedAceMasks, &hSecHeap);

		if (pSa == NULL)
			ReportError (_T ("Failure setting security attributes."), 3, TRUE);
		hFile = CreateFile (argv [FileIndex], 0,
				0, pSa, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile == INVALID_HANDLE_VALUE)
			ReportError (_T ("Failure creating secure file."), 4, TRUE);
		CloseHandle (hFile);
		HeapDestroy (hSecHeap); /* Release the security structures */
	}
	else if (Exists) {
				/* File exists; change permissions. */
		Change = ChangeFilePermissions (Mode, argv [FileIndex], AllowedAceMasks,
			DeniedAceMasks);
		if (!Change && !Force)
			ReportError (_T ("Change Permissions error."), 5, TRUE);
	}
	else /* Error - File does not exist and no -c option. */
		ReportError (_T ("File to change does not exist."), 6, 0);
	
	return 0;
}
